from ckantools.decorators import auth


@auth(anon=True)
def vds_download_queue(context, data_dict) -> dict:
    return {'success': True}


@auth(anon=True)
def vds_download_regenerate(context, data_dict) -> dict:
    return {'success': True}


@auth(anon=True)
def vds_get_avro_schema(context, data_dict) -> dict:
    return {'success': True}


@auth()
def vds_custom_download_filename(context, data_dict):
    # only allow access to admins (they usually skip this check)
    user_is_sysadmin = context.get('auth_user_obj').sysadmin
    return {'success': user_is_sysadmin}



from pathlib import Path
from typing import List, Optional, Tuple

from ckan.lib import search
from rq.job import Job
from splitgill.indexing.syncing import BulkOptions
from splitgill.model import IngestResult, Record

from ckanext.versioned_datastore.lib.importing.details import (
    create_details,
    get_last_file_hash,
)
from ckanext.versioned_datastore.lib.importing.ingest import (
    download_resource_data,
    iter_records,
)
from ckanext.versioned_datastore.lib.importing.readers import choose_reader_for_resource
from ckanext.versioned_datastore.lib.tasks import Task
from ckanext.versioned_datastore.lib.utils import (
    ReadOnlyResourceException,
    get_database,
    is_resource_read_only,
)
from ckanext.versioned_datastore.model.stats import INDEX, INGEST, PREP, ImportStats


def queue_ingest(
    resource: dict,
    replace: bool,
    api_key: str,
    records: Optional[List[dict]] = None,
) -> Tuple[Job, Job]:
    """
    Queues a new ingest job on the importing queue for the given resource.

    :param resource: the resource dict
    :param replace: whether any existing records should be deleted before adding the new
        records.
    :param api_key: the user's API key to use for file access (this is needed for
        private datasets)
    :param records: optional list of dicts to ingest instead of the file/URL on the
        resource
    :returns: a 2-tuple of the created ingest job and sync job which is dependent on the
        ingest job
    """
    if is_resource_read_only(resource['id']):
        raise ReadOnlyResourceException('This resource has been marked as read only')

    # create the ingest task first
    ingest_task = IngestResourceTask(resource, replace, api_key, records)
    ingest_job = ingest_task.queue()

    # then create a sync task dependent on the ingest task
    sync_task = SyncResourceTask(resource)
    sync_job = sync_task.queue(depends_on=ingest_job)

    return ingest_job, sync_job


def queue_delete(resource: dict) -> Tuple[Job, Job]:
    """
    Queues a new delete job for the given resource to remove all records. This is a
    versioned delete, so in reality all the data is maintained, just the latest versions
    of all records are set to empty.

    :param resource: the resource dict
    :returns: a 2-tuple of the created ingest job and sync job which is dependent on the
        ingest job
    """
    if is_resource_read_only(resource['id']):
        raise ReadOnlyResourceException('This resource has been marked as read only')

    # create the delete task first
    ingest_task = DeleteResourceTask(resource)
    delete_job = ingest_task.queue()

    # then create a sync task dependent on the delete task
    sync_task = SyncResourceTask(resource)
    sync_job = sync_task.queue(depends_on=delete_job)

    return delete_job, sync_job


def queue_sync(resource: dict, full: bool = False) -> Job:
    """
    Queues a sync job to synchronise any changes in MongoDB for this resource with
    Elasticsearch.

    :param resource: the resource dict
    :param full: whether to completely resync the Elasticsearch data with MongoDB or
        just sync the changes (default: False, just sync the changes)
    :returns: the queued sync job
    """
    if is_resource_read_only(resource['id']):
        raise ReadOnlyResourceException('This resource has been marked as read only')

    return SyncResourceTask(resource, full).queue()


def get_dupe_message(file_hash: str) -> str:
    """
    Returns an error message to be used for duplicate resource ingestions where nothing
    needs to be done. No error has really occurred so raising an exception feels like a
    bad way to deal with this, hence this str generator.

    :param file_hash: the file's hash
    :returns: a str to be stored as the stat's "error"
    """
    return f'This file has been ingested before, ignoring [hash: {file_hash}]'


class IngestResourceTask(Task):
    """
    Class representing a task to ingest a resource from a file/URL or from a given list
    of record dicts.
    """

    def __init__(
        self,
        resource: dict,
        replace: bool,
        api_key: str,
        records: Optional[List[dict]] = None,
    ):
        """
        :param resource: the resource dict
        :param replace: whether to replace all existing records with the records that
                        will be ingested or not
        :param api_key: the user's API key to use for file access (this is needed for
                        private datasets)
        :param records: optional list of dicts to ingest instead of the resource's
                        file/URL
        """
        self.resource = resource
        self.replace = replace
        self.api_key = api_key
        self.records = records

        # create a meaningful title
        if records is not None:
            data_info = f'{len(records)} records'
        else:
            data_info = 'data from file/url'
        title = f'Ingest for {self.resource_id} of {data_info} [replace: {replace}]'
        super().__init__('importing', title)

    @staticmethod
    def result_to_dict(result: IngestResult) -> dict:
        """
        Simple helper to convert an IngestResult object to a dict.

        :param result: the IngestResult object
        :returns: a dict
        """
        return {
            'inserted': result.inserted,
            'deleted': result.deleted,
            'updated': result.updated,
        }

    @property
    def resource_id(self) -> str:
        """
        :returns: the ID of the resource this ingest is operating on
        """
        return self.resource['id']

    def run(self, tmpdir: Path):
        """
        Performs the ingestion.

        :param tmpdir: a dir to use for temporary storage
        """
        # do the prep stage, this involves downloading the data
        with ImportStats.track(self.resource_id, PREP) as stats:
            if self.records is None:
                source = tmpdir / 'source'
                file_hash = download_resource_data(self.resource, source, self.api_key)
                last_hash = get_last_file_hash(self.resource_id)
                if file_hash == last_hash:
                    stats.update(error=get_dupe_message(file_hash))
                    self.log.info(get_dupe_message(file_hash))
                    return
            else:
                source = self.records
                file_hash = None

            reader = choose_reader_for_resource(self.resource, source)
            self.log.info(f'Using reader {reader.get_name()}')
            stats.update(count=reader.get_count())

        with ImportStats.track(self.resource_id, INGEST) as stats:
            database = get_database(self.resource_id)
            try:
                operations = {}
                if self.replace and database.has_data():
                    # to do a replace we delete everything before doing the ingest
                    replace_result = database.ingest(
                        (
                            Record.delete(record.id)
                            for record in database.iter_records()
                        ),
                        commit=False,
                    )
                    operations['replace'] = self.result_to_dict(replace_result)
                ingest_result = database.ingest(
                    iter_records(reader.read(), stats), commit=False
                )
                operations['ingest'] = self.result_to_dict(ingest_result)
                stats.update(
                    count=database.data_collection.count_documents({'version': None}),
                    operations=operations,
                )
            except:
                self.log.exception('Error while ingesting data, rolling back')
                database.rollback_records()
                raise

            version = database.commit()
            if version is None:
                self.log.info('No changes detected for data')
                stats.update(count=0)
                return
            else:
                stats.update(version=version)
                self.log.info(f'Ingested new data with version {version}')

            try:
                create_details(
                    self.resource_id, version, reader.get_fields(), file_hash
                )
            except Exception as e:
                self.log.warning(
                    f'Failed to create DatastoreResourceDetails due to {e}'
                )

        self.log.info('Finished ingesting')


class SyncResourceTask(Task):
    def __init__(
        self,
        resource: dict,
        full: bool = False,
    ):
        self.resource = resource
        self.full = full
        super().__init__('importing', f'Sync of {self.resource_id} [full: {self.full}]')

    @property
    def resource_id(self) -> str:
        return self.resource['id']

    def run(self, tmpdir: Path):
        with ImportStats.track(self.resource_id, INDEX) as stats:
            database = get_database(self.resource_id)
            index_version = database.get_elasticsearch_version()
            data_version = database.get_committed_version()
            # indicate that we've started doing something
            stats.update(count=0)

            if not self.full and index_version == data_version:
                # nothing to do
                self.log.info('Elasticsearch is already in sync, nothing to do')
            else:
                # work out how many records will be affected by the sync
                if self.full or index_version is None:
                    count = database.data_collection.count_documents({})
                else:
                    count = database.data_collection.count_documents(
                        {'version': {'$gt': index_version}}
                    )

                # use fairly modest values for syncing
                sync_options = BulkOptions(100, 2, 3)
                # do the sync and log/save info
                result = database.sync(bulk_options=sync_options, resync=self.full)
                stats.update(
                    operations={'deleted': result.deleted, 'indexed': result.indexed},
                    count=count,
                    version=database.get_elasticsearch_version(),
                )
                self.log.info(
                    f'Finished, indexed: {result.indexed}, deleted: {result.deleted}'
                )

        # refresh the data about this package in the solr search index to ensure that
        # the datastore_active flag is set correctly. The flag is actually set in the
        # plugin via a before_show resource hook so asking CKAN to refresh the package
        # will force it to rebuild the resource dict and thus call before_show and get
        # datastore_active set
        search.rebuild(package_id=self.resource['package_id'])


class DeleteResourceTask(Task):
    def __init__(self, resource: dict):
        self.resource = resource
        super().__init__('importing', f'Delete {self.resource_id}')

    @property
    def resource_id(self) -> str:
        return self.resource['id']

    def run(self, tmpdir: Path):
        with ImportStats.track(self.resource_id, INGEST) as stats:
            database = get_database(self.resource_id)
            result = database.ingest(
                (Record.delete(record.id) for record in database.iter_records()),
                commit=False,
            )
            count = database.data_collection.count_documents({'version': None})
            version = database.commit()
            if version is None:
                self.log.info('No changes detected for data or options')
                stats.update(count=0)
            else:
                self.log.info(f'Deleted all records at {version}')
                stats.update(
                    version=version, count=count, operations=result_to_dict(result)
                )



from ckan.authz import is_sysadmin
from ckantools.decorators import auth


@auth(anon=True)
def vds_multi_query(context, data_dict) -> dict:
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def vds_multi_count(context, data_dict) -> dict:
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def vds_multi_autocomplete_value(context, data_dict) -> dict:
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def vds_multi_autocomplete_field(context, data_dict) -> dict:
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def vds_multi_hash(context, data_dict) -> dict:
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def vds_multi_fields(context, data_dict) -> dict:
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def vds_multi_stats(context, data_dict) -> dict:
    # allow access to everyone
    return {'success': True}


@auth()
def vds_multi_direct(context, data_dict):
    # only allow for admins
    return {'success': is_sysadmin(context.get('user'))}



import re
from collections import defaultdict
from functools import total_ordering
from itertools import chain, islice
from typing import Dict, List, Set, Tuple, Union

from splitgill.indexing.fields import ParsedField

from ckanext.versioned_datastore.lib.importing.details import get_all_details


@total_ordering
class Group:
    def __init__(self, name: str):
        self.name: str = name
        self.fields: List[Tuple[ParsedField, str]] = []

    @property
    def resource_count(self) -> int:
        return len({resource_id for _, resource_id in self.fields})

    @property
    def record_count(self) -> int:
        # todo: this was sum for a bit, check this ok
        return max(field.count for field, _ in self.fields)

    @property
    def path_to_resource_id_map(self) -> Dict[str, List[str]]:
        result = defaultdict(set)
        for field, resource_id in self.fields:
            result[field.path].add(resource_id)
        return {field: list(resource_ids) for field, resource_ids in result.items()}

    def add(self, resource_id: str, field: ParsedField):
        self.fields.append((field, resource_id))

    def __eq__(self, other) -> bool:
        if isinstance(other, Group):
            return other.name == self.name
        return NotImplemented

    def __lt__(self, other) -> bool:
        if isinstance(other, Group):
            return (self.resource_count, self.record_count, self.name) < (
                other.resource_count,
                other.record_count,
                other.name,
            )
        return NotImplemented


class FieldGroups:
    def __init__(self, skip_ids: bool = True):
        self.skip_ids = skip_ids
        self.ignore_groups: Set[re.Pattern] = set()
        self.force_groups: List[str] = []
        self.groups: Dict[str, Group] = {}

    def ignore(self, group: Union[re.Pattern, str]):
        if isinstance(group, re.Pattern):
            self.ignore_groups.add(group)
        else:
            self.ignore_groups.add(re.compile(group, re.IGNORECASE))

    def force(self, group: str):
        if group not in self.force_groups:
            self.force_groups.append(group.lower())

    def add(self, resource_id: str, fields: List[ParsedField]):
        for field in fields:
            group = field.path.lower()
            self.groups.setdefault(group, Group(group)).add(resource_id, field)

    def force_resource_fields(self, resource_id, version):
        # retrieve the datastore resource details if there are any
        all_details = get_all_details(resource_id, up_to_version=version)
        if all_details:
            # the all_details variable is an OrderedDict in ascending version order. We
            # want to iterate in descending version order though so that we respect the
            # column order at the version we're at before respecting any data from
            # previous versions
            for details in reversed(all_details.values()):
                for field_name in details.get_columns():
                    self.force(field_name)

    def select(self, count: int) -> List[dict]:
        groups = []
        forced = [
            self.groups[group] for group in self.force_groups if group in self.groups
        ]
        for group in sorted(self.groups.values(), reverse=True):
            # process skips and ignores
            if self.skip_ids and group.name == '_id':
                continue
            if any(ignore.match(group.name) for ignore in self.ignore_groups):
                continue
            # skip these, we'll add them later
            if group in forced:
                continue
            groups.append(group)

        return [
            {
                'group': group.name,
                'count': group.resource_count,
                'records': group.record_count,
                'fields': group.path_to_resource_id_map,
                'forced': group in forced,
            }
            for group in islice(chain(forced, groups), count)
        ]



from ckantools.decorators import auth


@auth(anon=True)
def vds_schema_latest(context, data_dict) -> dict:
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def vds_schema_validate(context, data_dict) -> dict:
    # allow access to everyone
    return {'success': True}



from contextlib import contextmanager
from datetime import datetime, timezone
from traceback import format_exception_only
from typing import Optional, Union

from ckan import model
from ckan.model import DomainObject, Session, meta
from ckan.model.types import JsonDictType
from sqlalchemy import (
    BigInteger,
    Boolean,
    Column,
    DateTime,
    Float,
    Table,
    UnicodeText,
    desc,
)
from sqlalchemy.exc import InvalidRequestError

# options for the ImportStats.type field
PREP = 'prep'
INGEST = 'ingest'
INDEX = 'index'
ALL_TYPES = [PREP, INDEX, INGEST]

# this table stores general "statistics" about the ingest and index events that occur on
# resources. It is also used to figure out what versions have been ingested and to a
# certain extent indexed and it's therefore pretty important (it is used to avoid
# ingesting older versions for example).
import_stats_table = Table(
    'versioned_datastore_import_stats',
    meta.metadata,
    Column('id', BigInteger, primary_key=True),
    Column('resource_id', UnicodeText, nullable=False, index=True),
    # the type of operation
    Column('type', UnicodeText, nullable=False),
    # the version this operation created or is working on, this will be null for ingests
    Column('version', BigInteger),
    # the start datetime of the operation
    Column('start', DateTime),
    # the end datetime of the operation
    Column('end', DateTime),
    # how long the operation took in seconds
    Column('duration', Float),
    # whether the operation is in progress or whether it has completed
    Column('in_progress', Boolean),
    # if there was an error, this column is populated with the details
    Column('error', UnicodeText),
    # the number of records handled during the operation
    Column('count', BigInteger),
    # the detailed operation breakdown returned by splitgill
    Column('operations', JsonDictType),
)


class ImportStats(DomainObject):
    """
    Object for holding resource import stats.
    """

    @classmethod
    def get(cls, stats_id: int) -> Optional['ImportStats']:
        return Session.query(cls).get(stats_id)

    def update(self, **kwargs):
        for field, value in kwargs.items():
            setattr(self, field, value)
        try:
            self.save()
        except InvalidRequestError:
            self.commit()

    @classmethod
    def begin(cls, resource_id: str, stat_type: str):
        stats = cls(
            resource_id=resource_id,
            type=stat_type,
            in_progress=True,
            start=datetime.now(timezone.utc),
        )
        stats.save()
        return stats

    @classmethod
    @contextmanager
    def track(cls, resource_id: str, stat_type: str):
        stats = cls.begin(resource_id, stat_type)
        try:
            yield stats
        except Exception as e:
            stats.mark_error(e)
            raise
        finally:
            stats.finish()

    def mark_error(self, error: Union[Exception, str]):
        """
        Marks the ImportStats object as having finished with an error. Just the error
        message is stored against the ImportStats object. "in_progress", "duration" and
        "end" are also updated.

        :param error: the exception object or string message
        """
        if isinstance(error, Exception):
            str_error = str(format_exception_only(type(error), error)[-1].strip())
        else:
            str_error = error
        self.update(error=str_error)
        self.finish()

    def finish(self):
        end = datetime.now(timezone.utc)
        self.update(
            in_progress=False, end=end, duration=(end - self.start).total_seconds()
        )


meta.mapper(ImportStats, import_stats_table)


def get_all_stats(resource_id):
    """
    Retrieves and returns all the ImportStats from the database associated with the
    given resource. They are ordered by ID descending which will result in the newest
    results coming back first.

    :param resource_id: the id of the resource
    :returns: a Query object which can be iterated over to retrieve all the results
    """
    return list(
        model.Session.query(ImportStats)
        .filter(ImportStats.resource_id == resource_id)
        .order_by(desc(ImportStats.id))
    )


def get_last_ingest(resource_id):
    """
    Retrieve the last ingest stat object from the database, or None if there aren't any.

    :param resource_id: the resource id
    :returns: an ImportStats object or None
    """
    return (
        model.Session.query(ImportStats)
        .filter(ImportStats.resource_id == resource_id)
        .filter(ImportStats.type == INGEST)
        .order_by(ImportStats.version.desc())
        .first()
    )



import pytest
from elasticsearch_dsl import Search
from splitgill.model import Record

from ckanext.versioned_datastore.lib.query.search.query import DirectQuery
from ckanext.versioned_datastore.lib.query.search.request import SearchRequest
from ckanext.versioned_datastore.lib.utils import get_database


@pytest.mark.usefixtures('with_vds')
class TestSearchRequest:
    def test_add_param(self):
        index = 'test-1'
        req = SearchRequest(DirectQuery([index]))

        # this should fail because the index doesn't exist
        with pytest.raises(Exception):
            req.run()

        req.add_param('ignore_unavailable', True)
        # this should now succeed because of above parameter
        result = req.run()
        assert result.count == 0

    def test_indexes_normal(self):
        database = get_database('test-1')
        query = DirectQuery([database.name])
        req = SearchRequest(query)

        # no version provided, use latest
        assert req.indexes() == [database.indices.latest]

        # add a version, use wildcard
        query.version = 10
        assert req.indexes() == [database.indices.wildcard]

    def test_indexes_force_no_version(self):
        database = get_database('test-1')
        req = SearchRequest(DirectQuery([database.name]), force_no_version=True)
        assert req.indexes() == [database.indices.wildcard]

    def test_set_no_result(self):
        database = get_database('test-1')

        # add a record to the database
        database.ingest([Record.new({'a': 5})])
        database.sync()

        req = SearchRequest(DirectQuery([database.name]))
        result_1 = req.run()
        assert result_1.count == 1
        assert result_1.hits

        req.set_no_results()
        result_2 = req.run()
        assert result_2.count == 1
        assert not result_2.hits

    def test_get_safe_size(self):
        query = DirectQuery(['test-1'])
        assert SearchRequest(query).get_safe_size() == 100
        assert SearchRequest(query, size=10).get_safe_size() == 10
        assert SearchRequest(query, size=0).get_safe_size() == 0
        assert SearchRequest(query, size=-1).get_safe_size() == 0
        assert SearchRequest(query, size=1004).get_safe_size() == 1000

    def test_to_search(self):
        query = DirectQuery(['test-1'])
        req = SearchRequest(query, size=10)
        search = req.to_search()
        assert isinstance(search, Search)
        # we need this so it must be here!
        assert search._extra['track_total_hits']
        # should always be size + 1 when size is set
        assert search._extra['size'] == 11
        # todo: maybe add more tests for this?



from ckanext.versioned_datastore.lib.query.utils import (
    convert_small_or_groups,
    remove_empty_groups,
)


class TestQueryUtils(object):
    def test_converts_single_item_or_group(self):
        test_query = {
            'filters': {'or': [{'string_equals': {'fields': ['x'], 'value': 'y'}}]}
        }
        expected_query = {
            'filters': {'and': [{'string_equals': {'fields': ['x'], 'value': 'y'}}]}
        }
        output_query = convert_small_or_groups(test_query)
        assert output_query == expected_query

    def test_converts_nested_single_item_or_group(self):
        test_query = {'filters': {'and': [{'or': [{'exists': {'fields': ['x']}}]}]}}
        expected_query = {
            'filters': {'and': [{'and': [{'exists': {'fields': ['x']}}]}]}
        }
        output_query = convert_small_or_groups(test_query)
        assert output_query == expected_query

    def test_does_not_convert_multi_item_or_group(self):
        test_query = {
            'filters': {
                'or': [{'exists': {'fields': ['x']}}, {'exists': {'fields': ['y']}}]
            }
        }
        output_query = convert_small_or_groups(test_query)
        assert output_query == test_query

    def test_does_not_convert_empty_or_group(self):
        test_query = {'filters': {'or': []}}
        output_query = convert_small_or_groups(test_query)
        assert output_query == test_query

    def test_removes_empty_groups(self):
        test_query = {
            'filters': {'and': [{'or': []}, {'not': [{'exists': {'fields': ['x']}}]}]}
        }
        expected_query = {
            'filters': {'and': [{'not': [{'exists': {'fields': ['x']}}]}]}
        }
        output_query = remove_empty_groups(test_query)
        assert output_query == expected_query

    def test_removes_filters_if_root_empty(self):
        test_query = {'filters': {'and': []}}
        expected_query = {}
        output_query = remove_empty_groups(test_query)
        assert output_query == expected_query



import abc
import csv
import shutil
import zipfile
from pathlib import Path
from typing import Iterable, List, Union

import openpyxl
import xlrd
from cchardet import UniversalDetector

from ckanext.versioned_datastore.lib.common import (
    ALL_FORMATS,
    SV_FORMATS,
    XLS_FORMATS,
    XLSX_FORMATS,
    ZIP_FORMATS,
)


class ReaderNotFound(Exception):
    """
    Exception indicating that no reader could be found for the provided file format.
    """

    def __init__(self, fmt: str):
        super().__init__(f"No reader matched the format '{fmt}'")
        self.fmt = fmt


def choose_reader(resource_format: str, source: Union[Path, List[dict]]) -> 'Reader':
    """
    Chooses an appropriate reader for the provided source. If the source is a file, this
    uses the resource_format to choose the reader, if the source is a list of dicts,
    then the MemoryReader is used.

    :param resource_format: the resource format
    :param source: the source data, either a path to a file, or a list of dicts
    :returns: a Reader instance
    """
    if isinstance(source, Path):
        if resource_format in SV_FORMATS:
            return SVReader(source)
        if resource_format in XLS_FORMATS:
            return XLSReader(source)
        if resource_format in XLSX_FORMATS:
            return XLSXReader(source)
        if resource_format in ZIP_FORMATS:
            return ZipReader(source)
        raise ReaderNotFound(resource_format)
    else:
        return MemoryReader(source)


def choose_reader_for_resource(
    resource: dict, source: Union[Path, List[dict]]
) -> 'Reader':
    """
    Selects a reader for the given resource using the format primarily, and resource url
    secondarily.

    :param resource: the resource dict
    :param source: the source data, either a path to a file, or a list of dicts
    :returns: a Reader instance
    """
    # start off trying to use the format they have set/CKAN has inferred
    resource_format = resource.get('format', '')
    if not resource_format:
        # if that isn't available, try and get it from the resource url
        url = resource['url']
        if url:
            resource_format = url.rsplit('.', 1)[-1]
        # todo: could try detecting based on the file? e.g. using python-magic etc
    return choose_reader(resource_format.lower(), source)


def detect_encoding(source: Path) -> str:
    """
    Given a file, attempt to detect the character encoding it uses.

    :param source: the path to the file
    :returns: the character encoding
    """
    with source.open('rb') as f:
        detector = UniversalDetector()
        # feed the universal detector the entire file
        while True:
            chunk = f.read(8192)
            if chunk:
                detector.feed(chunk)
            else:
                detector.close()
                break

        encoding = detector.result['encoding']
        # if the detector failed to work out the encoding (unlikely) or if the
        # encoding it comes up with is ASCII, just default to UTF-8 (UTF-8 is a
        # superset of ASCII)
        if encoding is None or encoding == 'ASCII':
            encoding = 'utf-8'

    return encoding


class Reader(abc.ABC):
    """
    Abstract base class for reader implementations.
    """

    @abc.abstractmethod
    def get_name(self) -> str:
        """
        :returns: a name for the Reader instance (for logging)
        """
        ...

    @abc.abstractmethod
    def get_fields(self) -> List[str]:
        """
        Returns the list of the field names found in the source.

        :returns: the fields in the source this Reader is reading
        """
        ...

    @abc.abstractmethod
    def read(self) -> Iterable[dict]:
        """
        Actually reads the data from the source and yields dicts for each row found.

        :returns: yields dicts
        """
        ...

    @abc.abstractmethod
    def get_count(self) -> int:
        """
        Returns the number of rows in the source.

        :returns: an integer representing the number of rows
        """
        ...


class SVReader(Reader):
    """
    Class for reading csv and tsv files.
    """

    def __init__(self, source: Path):
        """
        :param source: the source file path
        """
        self.source = source
        self.encoding = detect_encoding(self.source)
        with self.source.open(encoding=self.encoding, newline='') as f:
            # instead of relying on people to correctly declare the dialect they are
            # using (from experience people are awful at this), sniff it for ourselves
            self.dialect = csv.Sniffer().sniff(f.read(1024))
        self._count = None

    def get_name(self) -> str:
        """
        Creates a name for this SVReader instance which represents the dialect and
        character encoding found (useful for debugging why someone's data wasn't
        ingested as expected).

        :returns: the name
        """
        dialect_info = ', '.join(
            f'{attr}: {str(getattr(self.dialect, attr))}'
            for attr in [
                'lineterminator',
                'quoting',
                'doublequote',
                'delimiter',
                'quotechar',
                'skipinitialspace',
            ]
        )
        dialect_info = dialect_info.encode('unicode_escape').decode('utf-8')
        return f'SV reader, encoding: {self.encoding}, dialect: [{dialect_info}]'

    def get_fields(self) -> List[str]:
        """
        Reads the first line of the source file and returns it as a list.

        :returns: the field names
        """
        with self.source.open(encoding=self.encoding, newline='') as f:
            return next(csv.reader(f, dialect=self.dialect))

    def read(self) -> Iterable[dict]:
        """
        Reads each line of the source file and yields each row as a dict.

        :returns: yields each row as a dict
        """
        with self.source.open(encoding=self.encoding, newline='') as f:
            yield from csv.DictReader(f, dialect=self.dialect)

    def get_count(self) -> int:
        if self._count is None:
            with self.source.open(encoding=self.encoding, newline='') as f:
                # count each row once, then take 1 off assuming the first row is the
                # header. Use max to avoid returning a negative value if the file is
                # just empty
                self._count = max(0, sum(1 for _ in f) - 1)
        return self._count


class XLSReader(Reader):
    """
    Reader for XLS files (i.e. old Excel spreadsheets).
    """

    def __init__(self, source: Path):
        """
        :param source: the file path to the source XLS file
        """
        self.source = source
        with self.source.open('rb') as f:
            book = xlrd.open_workbook(f.read())
            # todo: currently we don't deal with multisheeted spreadsheets, we just
            #       choose the first sheet and roll with it
            sheet = book.sheet_by_index(0)
            self.header = [str(cell.value) for cell in sheet.row(0)]
            self.rows = list(sheet.rows())[1:]

    def get_name(self) -> str:
        return 'XLS reader'

    def get_fields(self) -> List[str]:
        return self.header

    def read(self) -> Iterable[dict]:
        """
        Yields a dict for each row in the file's first sheet. We do some basic handling
        of types to ensure text, numbers, and booleans are converted correctly. Dates
        are not handled because the way dates are handled in XLS files is extremely
        complicated, and it's easier just to tell people not to do it, or use a string
        representation that Splitgill can parse.

        Empty field names, empty values, and values we can't convert are ignored.

        :returns: yields a dict per row
        """
        converters = {
            # value should be a str, just use it
            xlrd.XL_CELL_TEXT: lambda x: x,
            # value should be a float, just use it
            xlrd.XL_CELL_NUMBER: lambda x: x,
            # value should be a float and dates in xls files are crackers so just use it
            xlrd.XL_CELL_DATE: lambda x: x,
            # value should be an int, convert it to a bool
            xlrd.XL_CELL_BOOLEAN: bool,
        }
        for row in self.rows[1:]:
            yield {
                field: converters[cell.ctype](cell.value)
                for field, cell in zip(self.header, row)
                # ignore empty field names and cell types we don't want to handle
                if field and cell.ctype in converters
            }

    def get_count(self) -> int:
        return len(self.rows)


class XLSXReader(Reader):
    """
    Reader for new style Excel spreadsheets.
    """

    def __init__(self, source: Path):
        """
        :param source: the path to the XLSX file
        """
        self.source = source
        with self.source.open('rb') as f:
            workbook = openpyxl.load_workbook(f, read_only=True)
            # todo: currently we don't deal with multisheeted spreadsheets, we just
            #       choose the first sheet and roll with it
            all_rows = list(workbook.worksheets[0].rows)
            self.header = [str(cell.value) for cell in all_rows[0]]
            self.rows = all_rows[1:]

    def get_name(self) -> str:
        return 'XLSX reader'

    def get_fields(self) -> List[str]:
        return self.header

    def read(self) -> Iterable[dict]:
        """
        Yields the rows from the spreadsheet's first sheet as dicts. All type
        conversions are handled by openpyxl. Empty field names or empty cell values are
        ignored.

        :returns: a dict per row
        """
        for row in self.rows[1:]:
            yield {
                field: cell.value
                for field, cell in zip(self.header, row)
                # ignore empty field names and empty cells
                if field and cell.value is not None
            }

    def get_count(self) -> int:
        return len(self.rows)


class MemoryReader(Reader):
    """
    Reader for a list of dicts already in memory.
    """

    def __init__(self, source: List[dict]):
        """
        :param source: a list of dicts
        """
        self.source = source

    def get_name(self) -> str:
        return 'Memory reader'

    def get_fields(self) -> List[str]:
        """
        Return the fields present in the source dicts. The fields will be returned in
        the order they are found in the source from first dict to last. This ensures the
        ordering is representative of the source.

        :returns: a list of field names
        """
        fields = []
        for record_data in self.source:
            for field in record_data:
                if field not in fields:
                    fields.append(field)
        return fields

    def read(self) -> Iterable[dict]:
        yield from self.source

    def get_count(self) -> int:
        return len(self.source)


class NoCandidateFileFoundInZip(Exception):
    """
    Raised when there is no file found in the zip which we can ingest.
    """

    def __init__(self):
        super().__init__('No candidate file was found in the zip file')


class ZipReader(Reader):
    """
    A reader for zip files.

    This reader looks in the zip and inspects the files inside in alphabetical order
    until it finds a file it can read, at which point it creates a reader for the file
    and this reader is used to fulfill the abstract base class's requirements.
    """

    def __init__(self, source: Path):
        """
        :param source: the path to the zip file
        """
        with zipfile.ZipFile(source) as temp_zip:
            # sort the list of files so that we maintain a consistency when reading zips
            for name in sorted(temp_zip.namelist()):
                extension = name.rsplit('.', 1)[-1]
                if extension in ALL_FORMATS:
                    extracted_source = source.parent / 'zipped_source'
                    with extracted_source.open('wb') as s:
                        with temp_zip.open(name, 'r') as f:
                            shutil.copyfileobj(f, s)
                    try:
                        self.reader = choose_reader(extension, extracted_source)
                        break
                    except ReaderNotFound:
                        continue
            else:
                raise NoCandidateFileFoundInZip()

    def get_name(self) -> str:
        return f'Zip reader, using {self.reader.get_name()}'

    def get_fields(self) -> List[str]:
        return self.reader.get_fields()

    def read(self) -> Iterable[dict]:
        yield from self.reader.read()

    def get_count(self) -> int:
        return self.reader.get_count()



# !/usr/bin/env python
# encoding: utf-8
import io
import json
import os

import jsonschema
import pytest
from splitgill.indexing.fields import DocumentField
from splitgill.search import ALL_POINTS, keyword, number, text

from ckanext.versioned_datastore.lib.query.schema import schema_base_path
from ckanext.versioned_datastore.lib.query.schemas.v1_0_0 import v1_0_0Schema


class TestV1_0_0Translator(object):
    def test_validate_examples(self):
        schema = v1_0_0Schema()
        path = os.path.join(schema_base_path, schema.version, 'examples')
        for filename in os.listdir(path):
            with io.open(os.path.join(path, filename), 'r', encoding='utf-8') as f:
                schema.validate(json.load(f))

    @staticmethod
    def compare_query_and_search(query_in, query_out):
        schema = v1_0_0Schema()
        # validate the query!
        schema.validate(query_in)
        # check that the translated version is correct compared to the expected query
        assert schema.translate(query_in).to_dict() == query_out

    def test_translate_1(self):
        query = {}
        query_out = {'match_all': {}}
        self.compare_query_and_search(query, query_out)

    def test_translate_2(self):
        query_in = {'search': 'mollusca'}
        query_out = {
            'match': {DocumentField.ALL_TEXT: {'query': 'mollusca', 'operator': 'and'}}
        }
        self.compare_query_and_search(query_in, query_out)

    def test_translate_3(self):
        query_in = {
            'filters': {
                'and': [
                    {'string_equals': {'fields': ['genus'], 'value': 'helix'}},
                    {
                        'string_contains': {
                            'fields': ['higherGeography'],
                            'value': 'europe',
                        }
                    },
                ]
            }
        }
        query_out = {
            'bool': {
                'filter': [
                    {'term': {keyword('genus'): 'helix'}},
                    {
                        'match': {
                            text('higherGeography'): {
                                'query': 'europe',
                                'operator': 'and',
                            }
                        }
                    },
                ]
            }
        }
        self.compare_query_and_search(query_in, query_out)

    def test_translate_4(self):
        query_in = {
            'search': 'italy',
            'filters': {
                'and': [
                    {'string_equals': {'fields': ['genus'], 'value': 'helix'}},
                    {
                        'string_contains': {
                            'fields': ['higherGeography'],
                            'value': 'europe',
                        }
                    },
                ]
            },
        }
        query_out = {
            'bool': {
                'filter': [
                    {'term': {keyword('genus'): 'helix'}},
                    {
                        'match': {
                            text('higherGeography'): {
                                'query': 'europe',
                                'operator': 'and',
                            }
                        }
                    },
                ],
                'must': [
                    {
                        'match': {
                            DocumentField.ALL_TEXT: {
                                'query': 'italy',
                                'operator': 'and',
                            }
                        }
                    }
                ],
            }
        }
        self.compare_query_and_search(query_in, query_out)

    def test_translate_5(self):
        query_in = {
            'filters': {
                'and': [
                    {'string_equals': {'fields': ['genus'], 'value': 'helix'}},
                    {
                        'or': [
                            {
                                'string_contains': {
                                    'fields': ['higherGeography'],
                                    'value': 'italy',
                                }
                            },
                            {
                                'string_contains': {
                                    'fields': ['higherGeography'],
                                    'value': 'spain',
                                }
                            },
                            {
                                'string_contains': {
                                    'fields': ['higherGeography'],
                                    'value': 'portugal',
                                }
                            },
                        ]
                    },
                ]
            }
        }
        query_out = {
            'bool': {
                'filter': [
                    {'term': {keyword('genus'): 'helix'}},
                    {
                        'bool': {
                            'minimum_should_match': 1,
                            'should': [
                                {
                                    'match': {
                                        text('higherGeography'): {
                                            'query': 'italy',
                                            'operator': 'and',
                                        }
                                    }
                                },
                                {
                                    'match': {
                                        text('higherGeography'): {
                                            'query': 'spain',
                                            'operator': 'and',
                                        }
                                    }
                                },
                                {
                                    'match': {
                                        text('higherGeography'): {
                                            'query': 'portugal',
                                            'operator': 'and',
                                        }
                                    }
                                },
                            ],
                        }
                    },
                ]
            }
        }
        self.compare_query_and_search(query_in, query_out)

    def test_translate_6(self):
        query_in = {
            'filters': {
                'and': [
                    {'string_equals': {'fields': ['genus'], 'value': 'helix'}},
                    {
                        'number_range': {
                            'fields': ['year'],
                            'less_than': 2010,
                            'less_than_inclusive': True,
                            'greater_than': 2000,
                            'greater_than_inclusive': True,
                        }
                    },
                    {
                        'or': [
                            {
                                'string_contains': {
                                    'fields': ['higherGeography'],
                                    'value': 'italy',
                                }
                            },
                            {
                                'string_contains': {
                                    'fields': ['higherGeography'],
                                    'value': 'spain',
                                }
                            },
                            {
                                'string_contains': {
                                    'fields': ['higherGeography'],
                                    'value': 'portugal',
                                }
                            },
                        ]
                    },
                ]
            }
        }
        query_out = {
            'bool': {
                'filter': [
                    {'term': {keyword('genus'): 'helix'}},
                    {'range': {number('year'): {'gte': 2000, 'lte': 2010}}},
                    {
                        'bool': {
                            'minimum_should_match': 1,
                            'should': [
                                {
                                    'match': {
                                        text('higherGeography'): {
                                            'query': 'italy',
                                            'operator': 'and',
                                        }
                                    }
                                },
                                {
                                    'match': {
                                        text('higherGeography'): {
                                            'query': 'spain',
                                            'operator': 'and',
                                        }
                                    }
                                },
                                {
                                    'match': {
                                        text('higherGeography'): {
                                            'query': 'portugal',
                                            'operator': 'and',
                                        }
                                    }
                                },
                            ],
                        }
                    },
                ]
            }
        }
        self.compare_query_and_search(query_in, query_out)

    def test_translate_7(self):
        query_in = {
            'filters': {
                'and': [
                    {'string_equals': {'fields': ['genus'], 'value': 'helix'}},
                    {
                        'geo_point': {
                            'latitude': 51.4712,
                            'longitude': -0.9421,
                            'radius': 10,
                            'radius_unit': 'mi',
                        }
                    },
                ]
            }
        }
        query_out = {
            'bool': {
                'filter': [
                    {'term': {keyword('genus'): 'helix'}},
                    {
                        'geo_distance': {
                            DocumentField.ALL_POINTS: {
                                'lat': 51.4712,
                                'lon': -0.9421,
                            },
                            'distance': '10mi',
                        }
                    },
                ]
            }
        }
        self.compare_query_and_search(query_in, query_out)

    def test_translate_8(self):
        query_in = {'filters': {'and': [{'exists': {'fields': ['associatedMedia']}}]}}
        query_out = {'exists': {'field': 'data.associatedMedia'}}
        self.compare_query_and_search(query_in, query_out)

    def test_translate_9(self):
        query_in = {'filters': {'and': [{'exists': {'geo_field': True}}]}}
        query_out = {'exists': {'field': ALL_POINTS}}
        self.compare_query_and_search(query_in, query_out)

    def test_translate_10(self):
        country = 'Curaçao'
        multipolygon = v1_0_0Schema().geojson['country'][country]
        query_in = {'filters': {'and': [{'geo_named_area': {'country': country}}]}}
        query_out = {
            'geo_polygon': {
                ALL_POINTS: {
                    'points': [
                        {'lat': point[1], 'lon': point[0]}
                        for point in multipolygon[0][0]
                    ]
                }
            }
        }
        self.compare_query_and_search(query_in, query_out)

    def test_translate_11(self):
        a_square = [
            [102.0, 2.0],
            [103.0, 2.0],
            [103.0, 3.0],
            [102.0, 3.0],
            [102.0, 2.0],
        ]
        another_square = [
            [100.0, 0.0],
            [101.0, 0.0],
            [101.0, 1.0],
            [100.0, 1.0],
            [100.0, 0.0],
        ]
        a_hole = [[100.2, 0.2], [100.8, 0.2], [100.8, 0.8], [100.2, 0.8], [100.2, 0.2]]

        def to_points(points):
            return [{'lat': point[1], 'lon': point[0]} for point in points]

        query_in = {
            'filters': {
                'and': [
                    {
                        'geo_custom_area': [
                            # just a square
                            [a_square],
                            # a square with a square hole in it
                            [another_square, a_hole],
                        ]
                    }
                ]
            }
        }
        query_out = {
            'bool': {
                'minimum_should_match': 1,
                'should': [
                    {'geo_polygon': {ALL_POINTS: {'points': to_points(a_square)}}},
                    {
                        'bool': {
                            'filter': [
                                {
                                    'geo_polygon': {
                                        ALL_POINTS: {
                                            'points': to_points(another_square)
                                        }
                                    }
                                }
                            ],
                            'must_not': [
                                {
                                    'geo_polygon': {
                                        ALL_POINTS: {'points': to_points(a_hole)}
                                    }
                                }
                            ],
                        }
                    },
                ],
            }
        }
        self.compare_query_and_search(query_in, query_out)

    def test_translate_12(self):
        query_in = {
            'filters': {
                'not': [{'string_equals': {'fields': ['genus'], 'value': 'helix'}}]
            }
        }
        query_out = {'bool': {'must_not': [{'term': {keyword('genus'): 'helix'}}]}}
        self.compare_query_and_search(query_in, query_out)

    def test_translate_ignores_additional_properties_in_filters(self):
        schema = v1_0_0Schema()

        nope = {
            'filters': {
                'something_else': {},
                'not': [{'string_equals': {'fields': ['genus'], 'value': 'helix'}}],
            }
        }
        with pytest.raises(jsonschema.ValidationError):
            schema.validate(nope)

    def test_translate_ignores_additional_properties_in_geo_named_area(self):
        schema = v1_0_0Schema()

        nope = {
            'filters': {
                'and': [
                    {
                        'geo_named_area': {
                            'country': 'Belgium',
                            'something_else': False,
                        }
                    }
                ]
            }
        }
        with pytest.raises(jsonschema.ValidationError):
            schema.validate(nope)



from bisect import bisect_right
from typing import Optional

from splitgill.indexing.options import ParsingOptionsBuilder
from splitgill.model import ParsingOptions

from ckanext.versioned_datastore.lib.utils import get_database, ivds_implementations


def get_options(
    resource_id: str,
    version: Optional[int] = None,
) -> Optional[ParsingOptions]:
    """
    Retrieves the resource's parsing options that apply for the given version. If the
    version is None, returns the latest options.

    :param resource_id: the resource ID
    :param version: optional version which, if passed, will mean the options returned
        are the options that would be available at the given version.
    :returns: a ParsingOptions instance or None if no parsing options are in use
    """
    database = get_database(resource_id)
    all_options = database.get_options(include_uncommitted=False)
    if not all_options:
        return None

    versions = sorted(all_options.keys())
    if version is None:
        # no version provided, return the latest options
        return all_options[versions[-1]]
    else:
        # guard against the version requested being below the earliest options version
        if version < versions[0]:
            return None

        # figure out which options version is in use at the requested version
        rounded_version = versions[bisect_right(versions, version) - 1]
        return all_options[rounded_version]


def update_options(resource_id: str, overrides: Optional[dict] = None) -> Optional[int]:
    """
    Update the options for the given resource using the overrides if given. If no
    options are set on the resource then some default ones are created and added.

    :param resource_id: the resource ID
    :param overrides: any override options to set
    :returns: the new options version that has been committed or None if nothing was
        changed
    """
    overrides = overrides or {}
    current_options = get_options(resource_id)
    if current_options is None:
        builder = create_default_options_builder()
    else:
        builder = ParsingOptionsBuilder(based_on=current_options)

    if 'keyword_length' in overrides:
        builder.with_keyword_length(overrides['keyword_length'])
    if 'float_format' in overrides:
        builder.with_float_format(overrides['float_format'])
    for true_value in overrides.get('true_values', []):
        builder.with_true_value(true_value)
    for false_value in overrides.get('false_values', []):
        builder.with_false_value(false_value)
    for date_format in overrides.get('date_formats', []):
        builder.with_date_format(date_format)
    for geo_hint in overrides.get('geo_hints', []):
        builder.with_geo_hint(*geo_hint)

    for plugin in ivds_implementations():
        plugin.vds_update_options(resource_id, builder)

    database = get_database(resource_id)
    return database.update_options(builder.build(), commit=True)


def create_default_options_builder() -> ParsingOptionsBuilder:
    """
    Creates an options builder with some sensible default options set. This is
    deliberately pretty slim (just the minimum required options are set). Use
    vds_update_options to make modifications.

    :returns: a ParsingOptionsBuilder instance
    """
    builder = ParsingOptionsBuilder()
    builder.with_keyword_length(8191)
    builder.with_float_format('{0:.15g}')
    return builder



from unittest.mock import MagicMock, patch

import pytest
from ckan.plugins import toolkit

from ckanext.versioned_datastore.logic.download.arg_objects import QueryArgs
from tests.helpers import patches


class TestDownloadQuery:
    def test_create_query(self):
        check_mock = MagicMock(return_value=True)
        with patch(
            'ckanext.versioned_datastore.logic.validators.check_resource_id', check_mock
        ):
            with patches.query_schemas():
                q = QueryArgs(
                    query={},
                    query_version='v1.0.0',
                    resource_ids=['resource-id-here'],
                ).to_schema_query()
            assert q.query == {}
            assert q.query_version == 'v1.0.0'
            assert q.resource_ids == ['resource-id-here']

    # todo: skipped because ckantools might have a bug in it where if one validation
    #       error is raised, it tries to re-raise errors[0] but errors is a dict not a
    #       list (ckantools/validators/ivalidators.py#L87)
    @pytest.mark.skip
    def test_create_query_not_datastore(self):
        check_mock = MagicMock(return_value=False)
        with patch(
            'ckanext.versioned_datastore.logic.validators.check_datastore_resource_id',
            check_mock,
        ):
            with patches.query_schemas():
                with pytest.raises(
                    toolkit.Invalid, match='No resource IDs are datastore resources'
                ):
                    QueryArgs(
                        query={'resource_ids': ['resource-id-here']},
                        query_version='v1.0.0',
                    )

    @pytest.mark.usefixtures('with_vds')
    def test_create_query_from_query_args(self):
        resource_ids = ['test-resource-id']
        check_mock = MagicMock(return_value=True)
        with patch(
            'ckanext.versioned_datastore.logic.validators.check_datastore_resource_id',
            check_mock,
        ):
            query_args = QueryArgs(
                query={
                    'filters': {
                        'and': [
                            {
                                'string_equals': {
                                    'fields': ['collectionCode'],
                                    'value': 'bot',
                                }
                            }
                        ]
                    },
                    'resource_ids': resource_ids,
                }
            )
            q = query_args.to_schema_query()

        assert q.query_version == 'v1.0.0'
        assert q.resource_ids == resource_ids
        assert 'filters' in q.query
        assert q.hash == '83fc3087623dfd7371cf697d1de6c879de3e722a'



import re
from typing import Iterable, Optional, Set, TypeVar

from ckan import plugins
from ckan.plugins import get_plugin, toolkit
from elasticsearch import Elasticsearch
from pymongo import MongoClient
from splitgill.manager import SplitgillClient, SplitgillDatabase

from ckanext.versioned_datastore.interfaces import (
    IVersionedDatastore,
    IVersionedDatastoreDownloads,
    IVersionedDatastoreQuerySchema,
)
from ckanext.versioned_datastore.lib import common


def get_available_datastore_resources(
    ignore_auth: bool = False, user_id: str = ''
) -> Set[str]:
    """
    Simple wrapper around get_available_resources which provides a list of available
    datastore resources to the currently logged-in user.

    :param ignore_auth: whether to ignore authentication (default: False)
    :returns: a set of resource IDs
    """
    return get_available_resources(
        datastore_only=True, ignore_auth=ignore_auth, user_id=user_id
    )


def get_available_resources(
    datastore_only: bool, ignore_auth: bool = False, user_id: str = ''
) -> Set[str]:
    """
    Get a set of resource IDs that are available to the currently logged-in user and, if
    datastore_only is set to True, are datastore active. If no user is logged-in, all
    public datastore resource IDs are returned. The resource IDs are returned as a set
    to enable quick checking between a list of requested IDs and the list of available
    IDs.

    :param datastore_only: whether to only return resource IDs that are datastore active
    :param ignore_auth: whether to ignore authentication (default: False)
    :returns: a set of resource IDs
    """
    resource_ids = set()

    offset = 0
    action = toolkit.get_action('current_package_list_with_resources')

    while True:
        context = {}
        if ignore_auth:
            # unless ignore auth is passed, in which case pass that in the context
            context['ignore_auth'] = True
        else:
            context['user'] = user_id
        packages = action(context, {'offset': offset, 'limit': 100})
        if not packages:
            break
        for package in packages:
            # add the datastore active resources
            resource_ids.update(
                resource['id']
                for resource in package['resources']
                if not datastore_only or resource.get('datastore_active', False)
            )
        offset += len(packages)

    return resource_ids


def get_database(resource_id: str) -> SplitgillDatabase:
    """
    Retrieves a SplitgillDatabase object for the given resource ID. If the
    SplitgillClient on the VDS plugin isn't yet configured, an exception is raised.

    :param resource_id: the resource's ID
    :returns: a SplitgillDatabase
    """
    name = get_sg_name(resource_id)
    return sg_client().get_database(name)


def sg_client() -> SplitgillClient:
    """
    Retrieves a Splitgill client object. If Splitgill is not configured yet on the VDS
    plugin, an exception is raised.

    :returns: an SplitgillClient object
    """
    vds_plugin = get_plugin('versioned_datastore')
    if not vds_plugin.is_sg_configured:
        raise Exception('VDS plugin not configured yet')
    return vds_plugin.sg_client


def es_client() -> Elasticsearch:
    """
    Retrieves an Elasticsearch client for use on the in use cluster. If Splitgill is not
    configured yet on the VDS plugin, an exception is raised.

    :returns: an Elasticsearch object
    """
    vds_plugin = get_plugin('versioned_datastore')
    if not vds_plugin.is_sg_configured:
        raise Exception('VDS plugin not configured yet')
    return vds_plugin.elasticsearch_client


def mongo_client() -> MongoClient:
    """
    Retrieves a Mongo client for use on the in use database instance. If Splitgill is
    not configured yet on the VDS plugin, an exception is raised.

    :returns: an MongoClient object
    """
    vds_plugin = get_plugin('versioned_datastore')
    if not vds_plugin.is_sg_configured:
        raise Exception('VDS plugin not configured yet')
    return vds_plugin.mongo_client


def get_latest_version(resource_id) -> Optional[int]:
    """
    Retrieves the latest version of the given resource from the status index.

    :param resource_id: the resource's id
    :returns: the version or None if the resource isn't indexed
    """
    return get_database(resource_id).get_elasticsearch_version()


def get_sg_name(resource_id: str) -> str:
    """
    Adds the configured prefix to the start of the resource id to get the index name for
    the resource data in elasticsearch.

    :param resource_id: the resource id
    :returns: the resource's Splitgill database name
    """
    prefix = toolkit.config.get('ckanext.versioned_datastore.sg_prefix', '')
    return f'{prefix}{resource_id}'


def unprefix_sg_name(sg_name: str) -> str:
    """
    Removes the configured prefix from the start of the index name to get the resource
    id.

    :param sg_name: the Spitgill database name
    :returns: the resource's id
    """
    prefix = toolkit.config.get('ckanext.versioned_datastore.sg_prefix', '')
    return sg_name[len(prefix) :]


def unprefix_index_name(sg_index_name: str) -> str:
    """
    Extracts the resource ID from the given Splitgill index name by removing the
    Splitgill specific parts, plus removing the prefix (if one is configured). If the
    resource ID cannot be extracted, a ValueError is raised.

    :param sg_index_name: the Splitgill index name
    :returns: the resource's ID
    """
    # all indexes have data- at the start and -latest or -arc-# on the end
    regexes = [re.compile(r'data-(.*)-latest'), re.compile(r'data-(.*)-arc-[0-9]+')]
    for regex in regexes:
        match = regex.match(sg_index_name)
        if match:
            return unprefix_sg_name(match.group(1))
    raise ValueError(f'Failed to extract resource name from index: {sg_index_name}')


class ReadOnlyResourceException(toolkit.ValidationError):
    """
    Raised when a write operation of some variety is attempted on a resource which has
    been marked as read only.
    """

    pass


class RawResourceException(toolkit.ValidationError):
    """
    Raised when trying to ingest a resource that has been marked with "disable_parsing".
    """

    pass


def is_resource_read_only(resource_id: str) -> bool:
    """
    Loops through the plugin implementations checking if any of them want the given
    resource id to be read only.

    :returns: True if the resource should be treated as read only, False if not
    """
    return any(
        plugin.vds_is_read_only_resource(resource_id)
        for plugin in ivds_implementations()
    )


def is_datastore_resource(resource_id: str) -> bool:
    """
    Checks if any data has made it to Elasticsearch for this resource ID. Note that this
    only checks Elasticsearch, it doesn't check MongoDB, and is therefore intended to
    simply test if there is any searchable data for the resource.

    :param resource_id: the resource id
    :returns: True if the resource is a datastore resource, False if not
    """
    return get_database(resource_id).get_elasticsearch_version() is not None


def is_datastore_only_resource(resource_url: str) -> bool:
    """
    Checks whether the resource url is a datastore only resource url. When uploading
    data directly to the API without using a source file/URL the url of the resource
    will be set to "_datastore_only_resource" to indicate that as such. This function
    checks to see if the resource URL provided is one of these URLs. Note that we check
    a few different scenarios as CKAN has the nasty habit of adding a protocol onto the
    front of these URLs when saving the resource, sometimes.

    :param resource_url: the URL of the resource
    :returns: True if the resource is a datastore only resource, False if not
    """
    return (
        resource_url == common.DATASTORE_ONLY_RESOURCE
        or resource_url == f'http://{common.DATASTORE_ONLY_RESOURCE}'
        or resource_url == f'https://{common.DATASTORE_ONLY_RESOURCE}'
    )


def is_ingestible(resource: dict) -> bool:
    """
    Returns True if the resource can be ingested into the datastore and False if not. To
    be ingestible, the resource must either be a datastore only resource (signified by
    the url being set to _datastore_only_resource) or have a format that we can ingest
    (the format field on the resource is used for this, not the URL). If the url is
    None, False is returned. This is technically not possible due to a Resource model
    constraint, but it's worth covering off anyway.

    :param resource: the resource dict
    :returns: True if it is, False if not
    """
    if resource.get('url', None) is None:
        return False

    resource_format = resource.get('format', None)
    not_raw = not resource.get('disable_parsing', False)
    return (
        not_raw
        and is_datastore_only_resource(resource['url'])
        or (
            resource_format is not None
            and resource_format.lower() in common.ALL_FORMATS
        )
    )


T = TypeVar('T', bound=IVersionedDatastore)
U = TypeVar('U', bound=IVersionedDatastoreDownloads)
V = TypeVar('V', bound=IVersionedDatastoreQuerySchema)


def ivds_implementations() -> Iterable[T]:
    yield from plugins.PluginImplementations(IVersionedDatastore)


def idownload_implementations() -> Iterable[U]:
    yield from plugins.PluginImplementations(IVersionedDatastoreDownloads)


def iqs_implementations() -> Iterable[V]:
    yield from plugins.PluginImplementations(IVersionedDatastoreQuerySchema)



from ..validators import (
    ignore_missing,
    int_validator,
    not_empty,
    validate_datastore_resource_id,
)


def vds_version_schema() -> dict:
    return {}


def vds_version_record() -> dict:
    return {
        'resource_id': [not_empty, str, validate_datastore_resource_id],
        'record_id': [not_empty, str],
    }


def vds_version_resource() -> dict:
    return {
        'resource_id': [not_empty, str, validate_datastore_resource_id],
    }


def vds_version_round() -> dict:
    return {
        'resource_id': [not_empty, str, validate_datastore_resource_id],
        'version': [ignore_missing, int_validator],
    }



import dataclasses

from splitgill.indexing.fields import ParsedType


class InvalidSortException(Exception):
    def __init__(self, sort: 'Sort'):
        super().__init__(f'Cannot sort on {sort.type} field (field: {sort.field})')


@dataclasses.dataclass
class Sort:
    field: str
    ascending: bool = True
    type: ParsedType = ParsedType.KEYWORD

    @property
    def is_sort_on_id(self) -> bool:
        return self.field == '_id'

    def to_sort(self) -> dict:
        """
        Produce a sort dict for the Elasticsearch request.

        :returns: a dict
        """
        # you're not allowed to sort on text types
        if self.type == ParsedType.TEXT:
            raise InvalidSortException(self)

        unmapped_types = {
            ParsedType.NUMBER: 'double',
            ParsedType.DATE: 'date',
            ParsedType.BOOLEAN: 'boolean',
            ParsedType.KEYWORD: 'keyword',
        }

        options = {
            'order': 'asc' if self.ascending else 'desc',
            'missing': '_last',
            'unmapped_type': unmapped_types[self.type],
        }
        # TODO: do we need this?
        if self.type == ParsedType.DATE:
            options['format'] = 'epoch_millis'
        return {self.type.path_to(self.field, full=True): options}

    @classmethod
    def from_basic(cls, field_and_sort: str) -> 'Sort':
        if field_and_sort.endswith(' desc'):
            return Sort(field_and_sort[:-5], False)
        elif field_and_sort.endswith(' asc'):
            return Sort(field_and_sort[:-4], True)
        return Sort(field_and_sort, True)

    @classmethod
    def from_multi(cls, field: str, order: str, parsed_type: ParsedType) -> 'Sort':
        # only pass ascending=False if the order is desc, otherwise pass True
        return Sort(field, not order == 'desc', parsed_type)

    @classmethod
    def desc(cls, field: str, parsed_type: ParsedType = ParsedType.KEYWORD) -> 'Sort':
        return Sort(field, False, parsed_type)

    @classmethod
    def asc(cls, field: str, parsed_type: ParsedType = ParsedType.KEYWORD) -> 'Sort':
        return Sort(field, True, parsed_type)



from ckantools.decorators import auth


@auth(anon=True)
def vds_version_schema(context, data_dict) -> dict:
    return {'success': True}


@auth(anon=True)
def vds_version_resource(context, data_dict) -> dict:
    return {'success': True}


@auth(anon=True)
def vds_version_record(context, data_dict) -> dict:
    return {'success': True}


@auth(anon=True)
def vds_version_round(context, data_dict) -> dict:
    return {'success': True}



from operator import itemgetter
from typing import Dict, List, Optional

from splitgill.indexing.fields import DataField, ParsedField, ParsedType
from splitgill.search import keyword

from ckanext.versioned_datastore.lib.importing.details import get_details_at
from ckanext.versioned_datastore.lib.query.search.query import BasicQuery
from ckanext.versioned_datastore.lib.query.search.request import SearchRequest
from ckanext.versioned_datastore.lib.query.search.sort import Sort
from ckanext.versioned_datastore.lib.utils import get_database


def find_version(data_dict: dict) -> Optional[int]:
    """
    Retrieve the version from the data_dict. The version can be specified as a parameter
    in its own right or as a special filter in the filters dict using the key
    __version__. Using the version parameter is preferred and will override any filter
    version value. The filter method is provided because of limitations in the CKAN
    recline.js framework used by the NHM on CKAN 2.3 where no additional parameters can
    be passed other than q, filters etc.

    :param data_dict: the data dict, this might be modified if the __version__ key is
        used (it will be removed if present)
    :returns: the version found as an integer, or None if no version was found
    """
    version = data_dict.get('version')
    # pop the __version__ to avoid including it in the normal search filters, even if we
    # don't use it
    filter_version = data_dict.get('filters', {}).pop('__version__', None)

    if version is not None:
        return int(version)

    if filter_version is not None:
        # it'll probably be a list because it's a normal filter as far as the frontend
        # is concerned
        if isinstance(filter_version, list):
            # just use the first value
            filter_version = filter_version[0]
        if filter_version is not None:
            return int(filter_version)

    # no version found, return None
    return None


def make_request(data_dict: dict) -> SearchRequest:
    """
    Creates a SearchRequest from the given data_dict and returns it. The SearchRequest
    object will use a BasicQuery as its query which will be made by this function from
    the given data_dict. This function should be used by all basic actions which need to
    use basic queries for searches.

    :param data_dict: the data_dict passed to the action
    :returns: a SearchRequest object
    """
    query = BasicQuery(
        data_dict['resource_id'],
        find_version(data_dict),
        data_dict.get('q'),
        data_dict.get('filters'),
    )
    request = SearchRequest(
        query,
        size=data_dict.get('limit'),
        offset=data_dict.get('offset'),
        after=data_dict.get('after'),
        sorts=list(map(Sort.from_basic, data_dict.get('sort', []))),
        fields=data_dict.get('fields', []),
        data_dict=data_dict,
    )
    if 'facets' in data_dict:
        facet_limits = data_dict.get('facet_limits', {})
        for facet in data_dict['facets']:
            request.add_agg(
                facet,
                'terms',
                field=keyword(facet),
                size=facet_limits.get(facet, 10),
            )
    return request


def format_facets(aggs: dict) -> dict:
    """
    Formats the facet aggregation result into the format we require. Specifically we
    expand the buckets out into a dict that looks like this:

        {
            "facet1": {
                "details": {
                    "sum_other_doc_count": 34,
                    "doc_count_error_upper_bound": 3
                },
                "values": {
                    "value1": 1,
                    "value2": 4,
                    "value3": 1,
                    "value4": 2,
                }
            },
            etc
        }

    etc.

    :param aggs: the aggregation dict returned from splitgill/elasticsearch
    :returns: the facet information as a dict
    """
    facets = {}
    for facet, details in aggs.items():
        facets[facet] = {
            'details': {
                'sum_other_doc_count': details['sum_other_doc_count'],
                'doc_count_error_upper_bound': details['doc_count_error_upper_bound'],
            },
            'values': {
                value_details['key']: value_details['doc_count']
                for value_details in details['buckets']
            },
        }

    return facets


def infer_type(
    data_field: DataField, parsed_fields: Dict[str, ParsedField], threshold: float = 0.8
) -> str:
    """
    Infers the type of each of the given data fields, using the parsed fields to
    determine what type to assign. A type is assigned to a field if more than the
    threshold percentage of values in the field are of the parsed type.

    This function uses the parsed types instead of the data types for three reasons:
      - often the field data types are just strings, so we'd need to use the parsed
        types anyway to get a non-string type
      - the response to the get_fields function where this function is called is
        typically used in a search result and therefore the fields response may be used
        to create another search in which case the parsed types would have to be used
        as you can't search using the data types
      - there is no date field data type, whereas there is a date field parsed type

    In most cases, the parsed type returned by this function will match the data type of
    the field's values.

    :param data_field: the data field to infer for
    :param parsed_fields: a dict of parsed paths to parsed fields
    :param threshold: a threshold determining the percentage of the values in the field
                      which must be of the parsed type to be inferred as it
                      (default: 0.8, i.e., 80%)
    :returns: the name of the inferred type (one of string, number, date, boolean)
    """
    parsed_field = parsed_fields.get(data_field.parsed_path)

    # this should only ever happen for fields which are always nested objects
    if not parsed_field:
        return 'object'

    # check for dates first, then booleans, then numbers, and then default to string
    for parsed_type in (ParsedType.DATE, ParsedType.BOOLEAN, ParsedType.NUMBER):
        if parsed_field.type_counts[parsed_type] / parsed_field.count >= threshold:
            # just so happens that the recline types match the ParsedType names
            return parsed_type.name.lower()
    return 'string'


def get_fields(resource_id: str, version: Optional[int] = None) -> List[dict]:
    """
    Given a resource id, returns the fields that existed at the given version. If the
    version is None then the fields for the latest version are returned.

    The response format is important as it must match the requirements of reclineJS's
    field definitions. See
    http://okfnlabs.org/recline/docs/models.html#field
     for more    details.

    All fields are returned by default as string or array types. This is because we have
    the capability to allow searchers to specify whether to treat a field as other types
    when searching, and therefore we don't need to try and guess the type, and we can
    leave it to the user to know the type which won't cause problems like interpreting a
    field as a number when it shouldn't be (for example a barcode like '013655395').

    The fields are returned in either alphabetical order, or if we have the ingestion
    details for the resource at the required version then the order of the fields will
    match the order of the fields in the original source.

    :param resource_id: the resource's id
    :param version: the version of the data we're querying (default: None, which means
        latest)
    :returns: a list of dicts containing the field data
    """
    database = get_database(resource_id)
    data_fields = database.get_data_fields(version)
    parsed_fields = {field.path: field for field in database.get_parsed_fields(version)}

    fields = []
    seen = {'_id'}

    for field in data_fields:
        if field.parsed_path in seen:
            continue
        field_repr = {
            'id': field.parsed_path,
            'type': infer_type(field, parsed_fields),
            'sortable': True,
        }
        if field.children:
            field_repr['sortable'] = False
            if any(child.is_list_element for child in field.children):
                field_repr['type'] = 'array'
        seen.add(field.parsed_path)
        fields.append(field_repr)

    details = get_details_at(resource_id, version)
    if details is None:
        # no details, just order by alphabetical field name
        fields.sort(key=itemgetter('id'))
    else:
        # we have details, order the fields using the order of the columns in the
        # original source
        column_order = details.get_columns(validate=False)

        def key(f: dict) -> int:
            try:
                return column_order.index(f['id'])
            except ValueError:
                return len(column_order)

        fields.sort(key=key)

    # add the _id field to the start of the field list
    fields.insert(0, {'id': '_id', 'type': 'string'})

    return fields
